/**
 * Functions of fingerprinting
 */
package functions;


import java.util.*;
import java.util.Map.Entry;

import entities.ByteQueue;
import entities.Code;
import entities.SymbolMap;
import tools.Calculator;
import tools.Configurator;
import tools.Randomizer;

/**
 * @author Liang Wang
 *
 */
public class Fingerprinting {
	/**
	 * Fingerprinting F from dF
	 * @param dF a data matrix may contain fingerprints.
	 * @param map the symbol map.
	 * @param F the fingerprint
	 * @param epsilon the lower limit ratio of tested fingerprint redundancy (usually set to 0.3).
	 * @param sigma the upper limit mean of byte usage differentialsz(usually set to 90).
	 * @param tau upper limit size of the public part of a map (usually set to 1048576).
	 * @param r the lower limit of correlation coefficient (usually set to 0.6).
	 * @return Returns true if determining that dF does contain F, otherwise returns false.
	 */
	public static boolean fingerprinting(byte[] dF, SymbolMap map, byte[] F, double epsilon, int sigma, int tau, double r) {
		//validation
		try {
			if(dF.length < 1) {
				throw new Exception("Too short of dF!");
			}
			if(map == null) {
				throw new Exception("MapPublic cannot be null!");
			}
			if(F.length < 32) {
				throw new Exception("Too short of F!");
			}
			if(epsilon <= 0 || epsilon > 1) {
				throw new Exception("The epsilon must be in (0,1]!");
			}
			if(sigma < 1) {
				throw new Exception("The sigma must not be less than 1!");
			}
		}
		catch(Exception ex) {
			ex.printStackTrace();
		}
		
		boolean phi_test = true;
		boolean sigma_test = true;
		boolean num_test = true;
		
		//Solve for rho and eta from map  3.1
		int rho = map.pLength;
		int eta = map.hLength;
		
		//Verify dF  3.2-3.4
		if(dF.length > (tau * rho) / (F.length + rho)) { //不满足这个条件就很可能被篡改了
			//指纹冗余度误差的下限、表示一个指纹至少应该被检测到多少副本
			epsilon = epsilon * tau * rho / ((F.length + rho) * dF.length);
			sigma = sigma * ((F.length + rho) * dF.length) / (tau * rho);
		}
		
		//Calculate phi 3.6
		int phiFromDF = (int)Math.floor(eta * dF.length * 1d / (rho * F.length));
		int phiFromMap = (int)Math.floor(eta * map.getPs().size() * 1d / F.length);
		int phi = Math.min(phiFromDF, phiFromMap);

		/*****Forward strict verification，第一种
		 * 1.通过解析明码来收集暗码序列
		 * 2.将解析出来的暗码出于指纹的长度，得到这段暗码有多少个指纹
		 * 3.通过把这解析出来的指纹和用户的指纹进行皮尔逊相关系数比对
		****/
		int phiStar = 0;
		for(int i = 0; i < rho; i ++) { //3.7
			phiStar = 0;
			int tail = (dF.length - i) % rho;
			byte[] dFnoTail = Calculator.subArray(dF, i, dF.length - tail);
			ByteQueue P = ByteQueue.getByteQueue(dFnoTail, dFnoTail.length / rho, rho);
			ByteQueue T = ByteQueue.getByteQueue(eta);
			if(P != null) {
				while(P.size() > 0) {  //3.8 解析明码收集暗码序列，并加入T
					byte[] x = P.poll();
					Code h = map.gethByp(new Code(x));
					if (h == null){
						byte[] xb = new byte[x.length];
						for(int j = 0; j < x.length; j ++) {
							xb[i] = (byte)(((int)x[j] * (int)(-1)));
						}
						Code h1 = map.gethByp(new Code(xb));
						if(h1 != null) {
							T.offer(h1.getData());
						}
					}
					if(h != null) {
						T.offer(h.getData());
					}
				}
			}
			byte [] Th = T.getAllData();
			if(Th != null) {
				tail = (Th.length % F.length);
				if(tail > 0) { //3.9 去除尾部空格
					Th = Calculator.subArray(Th, 0, Th.length - tail);
				}//|F|指纹长度，算出里面大概有多少个指纹
				ByteQueue Tf = ByteQueue.getByteQueue(Th, Th.length / F.length, F.length);
				if(Tf != null) {
					while(Tf.size() > 0) {
						byte[] FStar = Tf.poll(); //查看的同时，也将这个元素从容器中删除掉
						double rff = Fingerprinting.correlation(FStar, F, F.length);
						if(Math.abs(rff) >= r) {
							phiStar ++;
						}
					}
				}
			}
			
			Configurator.phiStar = phiStar;
			
			if(phiStar >= epsilon * phi) { //3.14
				phi_test = true;
				Configurator.mean = 0;
//				break; //This option should be used when in test.
				return true; //This option should be used when in practice.
			}
			else{
				phi_test = false;
			}
		}


		/*****Backward probabilistic verification，第二种
		 * 先估计一个给定指纹的完整明码序列
		 * 比较该明码序列与待处理数据中每个字节的使用差异
		 * 字节使用差异策略
		 * dF的字节使用量
		 *****/
		HashMap<String, Integer> U = new HashMap<String, Integer>();
		for(int i = 0; i < dF.length; i ++) {
			String key = String.valueOf(dF[i]);
			if(U.containsKey(key)) {
				U.put(key, U.get(key).intValue() + 1);
			}
			else {
				U.put(key, 1);
			}
		}//dF是带指纹的数据，格式就是明码部分,统计的是明码部分，也就是统计明码字节数

		//Get a eta-partition of Fphi with the tail trimmed. 3.17
		byte[] Fphi = new byte[phi * F.length - ((phi * F.length) % eta)];
		for(int i = 0; i < phi; i ++) {
			Calculator.fillArray(Fphi, F, i * F.length, (i + 1) * F.length);
		}
		ByteQueue Feta = ByteQueue.getByteQueue(Fphi, Fphi.length / eta, eta);

		//Get possible  3.17-3.20，将指纹划分，然后根据指纹找到对应的明码，然后统计明码字节数
		byte[] dFStar = null;
		while(Feta.size() > 0) {
			ArrayList<Code> Ph = map.getPByh(new Code(Feta.poll()));//返回与同一暗码相关的明码集
			int rand = Randomizer.randomInt(0, Ph.size());//在明码集中随机取一个下标
			byte[] p = Ph.get(rand).getData();//得到下标对应的数值
			dFStar = Calculator.concatArrays(dFStar, p);//将这个值加入到dFStar
		}
		//Byte usage of dFStar 统计每个字节的利用率  3.21
		HashMap<String, Integer> Up = new HashMap<String, Integer>();
		for(int i = 0; i < dFStar.length; i ++) {
			String key = String.valueOf(dFStar[i]);
			if(Up.containsKey(key)) {
				Up.put(key, Up.get(key).intValue() + 1);
			}
			else {
				Up.put(key, 1);
			}
		}
		//Compare the usage between the elements from U and Up, respectively.
		double mean = 0d;
		for(Entry<String, Integer> entry : Up.entrySet()) {
			mean = mean + Math.abs(entry.getValue() - (U.get(entry.getKey()) == null ? 0 : U.get(entry.getKey())));
		}
		mean = mean / Up.size();
		Configurator.mean = Math.round(mean * 10000d) / 10000d;

		//下面做测试第三种

		HashMap<Integer, Integer> U1 = new HashMap<Integer, Integer>();
		for(int i = 0; i < dF.length; i ++) {
			int key = dF[i];
			if(U1.containsKey(key)) {
				U1.put(key, U1.get(key).intValue() + 1);
			}
			else {
				U1.put(key, 1);
			}
		}
		HashMap<Integer, Integer> Up1 = new HashMap<Integer, Integer>();
		for(int i = 0; i < dFStar.length; i ++) {
			int key = dFStar[i];
			if(Up1.containsKey(key)) {
				Up1.put(key, Up1.get(key).intValue() + 1);
			}
			else {
				Up1.put(key, 1);
			}
		}
		int num = 0;
		int i = 0;
		List keyU = new ArrayList();
		Iterator<Map.Entry<Integer , Integer>> iterator1 =U1.entrySet().iterator();
		Iterator<Map.Entry<Integer , Integer>> iterator2 =Up1.entrySet().iterator();
		while(iterator1.hasNext()) {
			Map.Entry entry = iterator1.next();
			int key = (int) entry.getKey();
			keyU.add(key);
			i++;
		}
		int j = 0;
		while(iterator2.hasNext()) {
			Map.Entry entry2 = iterator2.next();
			int key2 = (int) entry2.getKey();//1,2,3,3;
			int key3 = (int) keyU.get(j);//-1,-2,-3,-4;
			if (key3 != 0){
				if ((key2 / key3 != 1) && (key2 / key3 != -1)){
					num++;
				}
			}
			j++;
		}
		int len = j;

//		Mean judgment
		if(mean <= sigma && num/len < 0.02) { // Mean should not be greater than an threshold.
			sigma_test = true;
		}
		else {
			sigma_test = false;
		}

		/**第三种**/
//		HashMap<Integer, Integer> U1 = new HashMap<Integer, Integer>();
//		for(int i = 0; i < dF.length; i ++) {
//			int key = dF[i];
//			if(U1.containsKey(key)) {
//				U1.put(key, U1.get(key).intValue() + 1);
//			}
//			else {
//				U1.put(key, 1);
//			}
//		}
//		HashMap<Integer, Integer> Up1 = new HashMap<Integer, Integer>();
//		for(int i = 0; i < dFStar.length; i ++) {
//			int key = dFStar[i];
//			if(Up1.containsKey(key)) {
//				Up1.put(key, Up1.get(key).intValue() + 1);
//			}
//			else {
//				Up1.put(key, 1);
//			}
//		}
//		int num = 0;
//		int i = 0;
//		List keyU = new ArrayList();
//		Iterator<Map.Entry<Integer , Integer>> iterator1 =U1.entrySet().iterator();
//		Iterator<Map.Entry<Integer , Integer>> iterator2 =Up1.entrySet().iterator();
//		while(iterator1.hasNext()) {
//			Map.Entry entry = iterator1.next();
//			int key = (int) entry.getKey();
//			keyU.add(key);
//			i++;
//		}
//		int j = 0;
//		while(iterator2.hasNext()) {
//			Map.Entry entry2 = iterator2.next();
//			int key2 = (int) entry2.getKey();//1,2,3,3;
//			int key3 = (int) keyU.get(j);//-1,-2,-3,-4;
//			if (key3 != 0){
//				if ((key2 / key3 != 1) && (key2 / key3 != -1)){
//					num++;
//				}
//			}
//			j++;
//		}
//		int len = j;
//		if(num / len <= 0.1) { // Mean should not be greater than an threshold.
//			num_test = true;
//		}
//		else {
//			num_test = false;
//		}

		//Final judgment
		return phi_test || sigma_test;
	}
	/**
	 * Calculates a Pearson correlation coefficient for two signals
	 * @param a signal one
	 * @param b signal two
	 * @param n length of a signal
	 * @return [-1,1]. Returns -10 if there are some thing wrong in parameters.
	 */
	private static double correlation(byte[] s1, byte[] s2, int n) {
	    if(s1 == null || s2 == null || n <= 0) {
	    	return -10;
	    }
	    
		double sum_s12 = 0.0;
	    double sum_s1  = 0.0;
	    double sum_s2  = 0.0;
	    double sum_s1s1 = 0.0; //s1^2
	    double sum_s2s2 = 0.0; //s2^2
	    double pxy = 0.0;
	    double temp1 = 0.0;
	    double temp2 = 0.0;
	    
	    for(int i = 0; i < n; i++)
	    {
	    	sum_s12 += (int)s1[i] * (int)s2[i];
	        sum_s1 += (int)s1[i];
	        sum_s2 += (int)s2[i];
	        sum_s1s1 += (int)s1[i] * (int)s1[i]; 
	        sum_s2s2 += (int)s2[i] * (int)s2[i]; 
	    }
	    
	    temp1 = n * sum_s1s1 - sum_s1 * sum_s1;
	    temp2 = n * sum_s2s2 - sum_s2 * sum_s2;
	    
	    if((temp1 > -0.0001d && temp1 < 0.0001d) || (temp2 > -0.0001d && temp2 < 0.0001d) || (temp1 * temp2 <= 0)){
	    	return -10;
        }
	    
	    pxy = (n * sum_s12 - sum_s1 * sum_s2) / Math.sqrt(temp1 * temp2);
	    
	    return pxy;
	}
}
