/**
 * Functions related to symbol mapping, i.e., phi 1.
 */
package functions;


import java.util.Arrays;

import entities.ByteQueue;
import entities.Code;
import entities.SymbolMap;
import tools.Calculator;
import tools.Configurator;
import tools.Converter;
import tools.Randomizer;

public class SymbolMapping {
	/**
	 * 用给定的私钥和指纹创建数据矩阵的符号映射
	 * Create a symbol map of a data matrix with a given private key and a fingerprint.
	 * @param d data matrix.
	 * @param hSK hash of private key (in Hex string).
	 * @param F fingerprint (in bytes).
	 * @param gamma redundancy factor (usually set to 0.3).
	 * @param tau upper limit size of the public part of a map (usually set to 1048576).
	 * @param presTheta a preset theta (a nearest theta will be set). Leaves it <1, if a hSK-specified one is needed.
	 * @param presEta a preset eta. Leaves it <1, if a hSK-specified one is needed.<br> Note: presEta <= |F| must hold.
	 * @param lambda symbol length factor (usually set to 1024).
	 * @return A symbol map in connection with d.
	 */
	public static SymbolMap createSymbolMap(byte[] d, String hSK, byte[] F, double gamma, int tau, int presTheta, int presEta, int lambda) {
		//Validation，不满足条件直接退出
		try {
			if(d.length < 1) {
				throw new Exception("Too short of d!");
			}
			if(hSK.length() < 64) {
				throw new Exception("Too short of hSK!");
			}
			if(F.length < 32) {
				throw new Exception("Too short of F!");
			}
			if(gamma <= 0 || gamma > 1) {
				throw new Exception("The gamma must be within the range of (0,1]!");
			}
			if(tau < (F.length + ((Math.log(d.length)/Math.log(2)) + 1) / 8d)) {
				throw new Exception("Too small of tau!");
			}
		}
		catch(Exception ex) {
			ex.printStackTrace();
		}

		//Get symbol length.1.1
		int theta = 0;
		if(presTheta > 0) {
			theta = Calculator.findNearestInt(d.length, presTheta);
		}
		else {
			theta = ParameterMaker.getSymbolLength(hSK, F.length, d.length, tau, lambda);
		}
		Configurator.theta = theta;

		//1.2、先补0，在切分theta份
		//Right zero padding of d.
		int x = d.length % theta;
		if(x > 0) {
			byte[] filler = new byte[theta - x];
			Arrays.fill(filler, (byte)0);
			d = Calculator.concatArrays(d, filler);
		}
		//Divide d by theta all in bytes.
		byte[][] S = new byte[d.length/theta][theta];
		for(int i = 0; i < d.length/theta; i ++) {
			for(int j = i * theta; j < (i + 1) * theta; j ++) {
				S[i][j - i * theta] = d[j];
			}
		}

		//Calculates plain code length: rho 1.3
		int rho = (int)Math.ceil(((Math.log(S.length) / Math.log(2)) + 1) / 8d);
		Configurator.rho = rho;

		//Get hidden code length: eta 1.4
		int eta = 0;
		if(presEta > 0 && presEta < F.length + 1) {
			eta = presEta;
		}
		else {
			eta = ParameterMaker.getHiddenCodeLength(hSK, S.length, F.length, gamma);
		}
		Configurator.eta = eta;

		//Calculate fingerprint redundancy: phi 1.5
		int phi = (int)Math.ceil(eta * S.length * 1d / F.length);
		Configurator.phi = phi;

		//Get redundant fingerprint and cut the tail. 1.6
		byte[] Fphi = new byte[eta * S.length];
		for(int i = 0; i < phi; i ++) {
			Calculator.fillArray(Fphi, F, i * F.length, (i + 1) * F.length);
		}
		ByteQueue Feta = ByteQueue.getByteQueue(Fphi, S.length, eta);

		//Create a coding space.1.7-1.11
		ByteQueue C = ByteQueue.getByteQueue(rho); //rho cannot be greater than 3!!!
		for(int i = 0; i < (int)Math.pow(2, 8 * rho); i ++) {
			String hex = Integer.toHexString(i);
			int comp = 2 * rho - hex.length();
			for(int j = 0; j < comp; j ++) {
				hex = "0" + hex;
			}
			byte[] b = Converter.Hex2byte(hex);
			C.offer(b);
		}

		//Generate a basic symbol map
		SymbolMap map = new SymbolMap();
		for(int i = 0; i < S.length; i ++) {
			int rand = Randomizer.randomInt(0, C.size());
			byte[] p_rho = C.remove(rand);
			map.addItem(new Code(S[i]), new Code(p_rho), new Code(Feta.element(i)), true);//分离map-、map+
		}

		//Obfuscation. 1.12
		int chi = ParameterMaker.getCodingRedundancy(hSK, map.getMapP().size(), rho);
		Configurator.chi = chi;

		//向其中加入chi个冗余的编码组，来混淆明码和映射的关系
		for(int i = 0; i < chi; i ++) { //1.13-1.18
			byte[] s = S[Randomizer.randomInt(0, S.length)];
			byte[] h = Feta.element(Randomizer.randomInt(0, Feta.size()));
			int rand =Randomizer.randomInt(0, C.size());
			byte[] p_rho = C.remove(rand);
			map.addItem(new Code(s), new Code(p_rho), new Code(h), false);
		}
		return map;
	}
}
